"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RouterClass = void 0;
const tslib_1 = require("tslib");
const mobx_1 = require("mobx");
const multikey_map_1 = tslib_1.__importDefault(require("multikey-map"));
const _schema_1 = require("../@schema");
const _state_1 = require("../@state");
const _utils_1 = require("../@utils");
const route_1 = require("../route");
const router_back_1 = require("./router-back");
class RouterClass {
    constructor(
    /** @internal */
    _schemas, 
    /** @internal */
    _options) {
        Object.defineProperty(this, "_schemas", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: _schemas
        });
        Object.defineProperty(this, "_options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: _options
        });
        /** @internal */
        Object.defineProperty(this, "_active", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /** @internal */
        Object.defineProperty(this, "_transition", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /** @internal */
        Object.defineProperty(this, "_switching", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /** @internal */
        Object.defineProperty(this, "_queue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        /** @internal */
        Object.defineProperty(this, "_stateInputMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new multikey_map_1.default()
        });
        (0, mobx_1.makeObservable)(this);
        for (let [key, childSchema] of Object.entries(_schemas)) {
            if (key.startsWith('$')) {
                continue;
            }
            if (childSchema === true) {
                childSchema = {};
            }
            this[key] = (0, route_1.createRoute)(this, childSchema, [key], new Map());
        }
    }
    get $path() {
        var _a;
        return (_a = this._active) === null || _a === void 0 ? void 0 : _a.entry.path;
    }
    get $state() {
        var _a;
        return (_a = this._active) === null || _a === void 0 ? void 0 : _a.entry.mergedState;
    }
    get $snapshot() {
        const activeEntry = this._active;
        if (!activeEntry) {
            return undefined;
        }
        const stateInputMap = this._stateInputMap;
        const schemas = this._schemas;
        const { operation, entry: { target }, } = activeEntry;
        const stateObjects = [];
        return {
            operation,
            entry: buildEntry(target),
            states: stateObjects.map(stateObject => (0, mobx_1.toJS)(stateObject)),
        };
        function buildEntry({ path, stateMap, previous, }) {
            const inputs = [];
            const states = [];
            let upperSchemas = schemas;
            for (const [index, key] of path.entries()) {
                const schema = (0, _schema_1.getChildSchema)(upperSchemas, key);
                const state = stateMap.get(index);
                const { $state: schemaState } = schema;
                if (typeof schemaState === 'function') {
                    const input = stateInputMap.get([state, schemaState]);
                    inputs.push(input);
                    states.push(-1);
                }
                else {
                    inputs.push(undefined);
                    const stateIndex = stateObjects.indexOf(state);
                    if (stateIndex >= 0) {
                        states.push(stateIndex);
                    }
                    else {
                        states.push(stateObjects.push(state) - 1);
                    }
                }
                upperSchemas = schema;
            }
            return {
                path,
                inputs,
                states,
                previous: previous && buildEntry(previous),
            };
        }
    }
    get $back() {
        const { entry: { previous: entry }, } = this._requireActive();
        if (!entry) {
            throw new Error('No previous entry');
        }
        return (0, router_back_1.createRouterBack)(this, entry);
    }
    get $ableToBack() {
        var _a;
        return ((_a = this._active) === null || _a === void 0 ? void 0 : _a.entry.previous) !== undefined;
    }
    $backTo(route) {
        const routes = Array.isArray(route) ? route : [route];
        let { entry: { previous: entry }, } = this._requireActive();
        outer: while (entry) {
            for (const route of routes) {
                if ((0, _utils_1.isArrayStartedWith)(entry.path, route.$path)) {
                    break outer;
                }
            }
            entry = entry.previous;
        }
        return entry ? (0, router_back_1.createRouterBack)(this, entry) : undefined;
    }
    $restore({ operation, entry, states }) {
        if (this._transition || this._switching) {
            throw new Error('Cannot restore during transition or switching');
        }
        const observableStates = states.map(state => (0, mobx_1.observable)(state));
        const stateInputMap = this._stateInputMap;
        const schemas = this._schemas;
        const target = restoreTarget(entry);
        (0, mobx_1.runInAction)(() => {
            this._active = {
                operation,
                entry: new route_1.RouteEntry(this, target.path, target.stateMap, target.previous, undefined),
            };
        });
        function restoreTarget(entry) {
            const { path, 
            // Adding defaults for compatibility
            inputs = [], states: stateIndexes, previous, } = entry;
            const stateMap = new Map();
            let upperSchemas = schemas;
            for (const [pathIndex, stateIndex] of stateIndexes.entries()) {
                const key = path[pathIndex];
                const schema = (0, _schema_1.getChildSchema)(upperSchemas, key);
                const { $state: schemaState } = schema;
                if (typeof schemaState === 'function') {
                    const input = inputs[pathIndex];
                    const state = schemaState(input, (0, _state_1.createMergedState)(stateMap));
                    (0, _state_1.assertState)(state, key);
                    const observableState = (0, mobx_1.observable)(state);
                    stateInputMap.set([observableState, schemaState], input);
                    stateMap.set(pathIndex, observableState);
                }
                else {
                    const state = stateIndex >= 0 ? observableStates[stateIndex] : undefined;
                    if (!state) {
                        throw new TypeError('State missing in snapshot');
                    }
                    stateMap.set(pathIndex, state);
                }
                upperSchemas = schema;
            }
            return {
                path,
                stateMap,
                previous: previous && restoreTarget(previous),
                statePart: undefined,
            };
        }
    }
    /** @internal */
    _reset(path, stateMapUpdate) {
        return (0, route_1.createRouteOperation)(this, 'reset', () => {
            var _a;
            const stateMap = this._buildStateMap(path, stateMapUpdate, (_a = this._active) === null || _a === void 0 ? void 0 : _a.entry);
            return {
                path,
                stateMap,
                previous: undefined,
            };
        });
    }
    /** @internal */
    _push(path, stateMapUpdate) {
        return (0, route_1.createRouteOperation)(this, 'push', () => {
            const { entry } = this._requireActive();
            const stateMap = this._buildStateMap(path, stateMapUpdate, entry);
            return {
                path,
                stateMap,
                previous: entry.target,
            };
        });
    }
    /** @internal */
    _replace(path, stateMapUpdate) {
        return (0, route_1.createRouteOperation)(this, 'replace', () => {
            const { entry } = this._requireActive();
            const stateMap = this._buildStateMap(path, stateMapUpdate, entry);
            return {
                path,
                stateMap,
                previous: entry.previous,
            };
        });
    }
    /** @internal */
    _switch(operation, { path, stateMap, previous, statePart }, switchingState) {
        const ref = {};
        if (switchingState === undefined) {
            switchingState = this._options.defaultSwitchingState;
            if (switchingState === undefined) {
                throw new Error('Switching state is required as no default is provided');
            }
        }
        const switchingStateObservable = mobx_1.observable.box(switchingState);
        const entry = new route_1.RouteEntry(this, path, stateMap, previous, statePart);
        (0, mobx_1.runInAction)(() => {
            this._switching = {
                operation,
                entry,
                switchingStateObservable,
                ref,
            };
        });
        return (0, route_1.createRouteSwitching)(this, switchingStateObservable, ref);
    }
    /** @internal */
    _completeSwitching(ref) {
        const { operation, entry } = this._requireSwitching(ref);
        (0, mobx_1.runInAction)(() => {
            entry.mergePendingStatePart();
            this._active = {
                operation,
                entry,
            };
            this._switching = undefined;
        });
    }
    /** @internal */
    _abortSwitching(ref) {
        this._requireSwitching(ref);
        (0, mobx_1.runInAction)(() => {
            this._switching = undefined;
        });
    }
    /** @internal */
    _set(operation, targetBuilder) {
        return {
            $completed: new Promise(resolve => this._startTransition(operation, targetBuilder, resolve)),
        };
    }
    /** @internal */
    _requireActive() {
        const active = this._active;
        if (!active) {
            throw new Error('No active entry');
        }
        return active;
    }
    /** @internal */
    _requireTransition() {
        const transition = this._transition;
        if (!transition) {
            throw new Error('No transition entry');
        }
        return transition;
    }
    /** @internal */
    _requireSwitching(ref) {
        const switching = this._switching;
        if (!switching) {
            throw new Error('No switching entry');
        }
        if (ref !== switching.ref) {
            throw new Error('Switching does not match');
        }
        return switching;
    }
    _startTransition(operation, targetBuilder, complete) {
        var _a;
        if (this._transition) {
            this._queue.push([operation, targetBuilder, complete]);
            return;
        }
        const target = targetBuilder();
        const activeRouteEntry = (_a = this._active) === null || _a === void 0 ? void 0 : _a.entry;
        const transitionRouteEntry = new route_1.RouteEntry(this, target.path, target.stateMap, target.previous, target.statePart);
        (0, mobx_1.runInAction)(() => {
            this._transition = {
                operation,
                entry: transitionRouteEntry,
            };
        });
        // Wait initial render to register transition.
        void Promise.resolve()
            .then(() => (0, mobx_1.when)(() => (!activeRouteEntry || !activeRouteEntry.blockedByLeaving) &&
            !transitionRouteEntry.blockedByEntering))
            .then(() => {
            this._finishTransition();
            complete();
        });
    }
    _finishTransition() {
        const { operation, entry } = this._requireTransition();
        (0, mobx_1.runInAction)(() => {
            entry.mergePendingStatePart();
            this._active = {
                operation,
                entry,
            };
            this._transition = undefined;
            const next = this._queue.shift();
            if (next) {
                const [operation, targetBuilder, complete] = next;
                this._startTransition(operation, targetBuilder, complete);
            }
        });
    }
    _buildStateMap(path, stateMapUpdate, active) {
        const { path: activePath, stateMap: activeStateMap } = active !== null && active !== void 0 ? active : {
            path: [],
            stateMap: new Map(),
        };
        const stateInputMap = this._stateInputMap;
        const stateMap = new Map();
        const commonStartKeys = (0, _utils_1.getCommonStartOfTwoArray)(path, activePath);
        let upperSchemas = this._schemas;
        for (const [index, key] of commonStartKeys.entries()) {
            const schema = (0, _schema_1.getChildSchema)(upperSchemas, key);
            if (stateMapUpdate.has(index)) {
                let state;
                let stateFunctionAndInput;
                if ('$state' in schema) {
                    // It is legit for this to be undefined in case of $state is a function
                    // accepting no argument or undefined argument.
                    const stateUpdate = stateMapUpdate.get(index);
                    const schemaState = schema.$state;
                    if (typeof schemaState === 'function') {
                        state = schemaState(stateUpdate, (0, _state_1.createMergedState)(stateMap));
                        stateFunctionAndInput = [schemaState, stateUpdate];
                    }
                    else {
                        state = stateUpdate;
                    }
                }
                (0, _state_1.assertState)(state, key);
                const observableState = (0, mobx_1.observable)(state);
                if (stateFunctionAndInput) {
                    const [stateFunction, stateInput] = stateFunctionAndInput;
                    stateInputMap.set([observableState, stateFunction], stateInput);
                }
                stateMap.set(index, observableState);
            }
            else {
                stateMap.set(index, activeStateMap.get(index));
            }
            upperSchemas = schema;
        }
        for (const [index, key] of path
            .slice(commonStartKeys.length)
            .entries()) {
            const schema = (0, _schema_1.getChildSchema)(upperSchemas, key);
            const stateIndex = commonStartKeys.length + index;
            let state;
            let stateFunctionAndInput;
            if ('$state' in schema) {
                const schemaState = schema.$state;
                if (stateMapUpdate.has(stateIndex)) {
                    const stateUpdate = stateMapUpdate.get(stateIndex);
                    if (typeof schemaState === 'function') {
                        state = schemaState(stateUpdate, (0, _state_1.createMergedState)(stateMap));
                        stateFunctionAndInput = [schemaState, stateUpdate];
                    }
                    else {
                        state = stateUpdate;
                    }
                }
                else {
                    if (typeof schemaState === 'function') {
                        state = schemaState(undefined, (0, _state_1.createMergedState)(stateMap));
                    }
                    else {
                        state = schemaState;
                    }
                }
            }
            else {
                state = {};
            }
            (0, _state_1.assertState)(state, key);
            const observableState = (0, mobx_1.observable)(state);
            if (stateFunctionAndInput) {
                const [stateFunction, stateInput] = stateFunctionAndInput;
                stateInputMap.set([observableState, stateFunction], stateInput);
            }
            stateMap.set(stateIndex, observableState);
            upperSchemas = schema;
        }
        return stateMap;
    }
}
tslib_1.__decorate([
    mobx_1.observable.ref
], RouterClass.prototype, "_active", void 0);
tslib_1.__decorate([
    mobx_1.observable.ref
], RouterClass.prototype, "_transition", void 0);
tslib_1.__decorate([
    mobx_1.observable.ref
], RouterClass.prototype, "_switching", void 0);
exports.RouterClass = RouterClass;
//# sourceMappingURL=router.js.map